/********************************************************************************
 * Copyright (c) 2023 CEA-List
 *
 * This program and the accompanying materials are made available under the
 * terms of the Eclipse Public License 2.0 which is available at
 * http://www.eclipse.org/legal/epl-2.0.
 *
 * SPDX-License-Identifier: EPL-2.0
 *
 ********************************************************************************/

#ifndef AIDGE_CORE_OPERATOR_OPERATOR_H_
#define AIDGE_CORE_OPERATOR_OPERATOR_H_

#include <memory>
#include <string>
#include <vector>

#include "aidge/backend/OperatorImpl.hpp"
#include "aidge/data/Data.hpp"
#include "aidge/data/Tensor.hpp"
#include "aidge/utils/Types.h"
#include "aidge/hook/Hook.hpp"

namespace Aidge {

class Operator : public std::enable_shared_from_this<Operator> {
protected:
    std::shared_ptr<OperatorImpl> mImpl; // implementation of the operator
    std::map<std::string, std::shared_ptr<Hook>> mHooks;

private:
    std::string mType;
    const IOIndex_t mNbData;
    const IOIndex_t mNbAttr;
    const IOIndex_t mNbOut;

public:
    Operator() = delete;
    Operator(const char* type, const IOIndex_t nbData, const IOIndex_t nbAttr, const IOIndex_t nbOut)
    : mType(type),
      mNbData(nbData),
      mNbAttr(nbAttr),
      mNbOut(nbOut)
    {
        // ctor
    }
    virtual std::shared_ptr<Operator> clone() const = 0;
    virtual ~Operator();

    Operator(const Operator& op):
        std::enable_shared_from_this<Operator>(),
        mNbData(op.mNbData),
        mNbAttr(op.mNbAttr),
        mNbOut(op.mNbOut)
    {
        mType = op.mType;
        mImpl = nullptr;
        // Implementation is never cloned. It is up to the non-abstract Operator copy-constructor to create a new implementation matching the copied Operator implementation.
        // See https://gitlab.eclipse.org/eclipse/aidge/aidge_core/-/merge_requests/8#note_1214050 for the discussion.
        // Hooks are not copied.
    }

public:

    virtual void associateInput(const IOIndex_t inputIdx, const std::shared_ptr<Data>* data) = 0;

    virtual std::shared_ptr<Data> getRawInput(const IOIndex_t inputIdx) const = 0;
    virtual std::shared_ptr<Data> getRawOutput(const IOIndex_t outputIdx) const = 0;

    std::shared_ptr<Hook> getHook(std::string hookName) {
        return mHooks[hookName];
    }
    void addHook(std::string hookName) {
        mHooks.insert(std::pair<std::string, std::shared_ptr<Hook>>(hookName,Registrar<Hook>::create({hookName})(shared_from_this())));
    }

    void runHooks() const;

///////////////////////////////////////////////////////
//        IMPLEMENTATION
///////////////////////////////////////////////////////

    virtual void setBackend(const std::string& name) = 0;
    virtual void setDatatype(const DataType& datatype) = 0;

    /**
     * @brief Set the a new OperatorImpl to the Operator
     *
     */
    void setImpl(std::shared_ptr<OperatorImpl> impl){
        mImpl = impl;
    }

    /**
     * @brief Minimum amount of data from a specific input for one computation pass.
     * @param inputIdx Index of the input analysed.
     * @return NbElts_t
     */
    virtual NbElts_t getNbRequiredData(const IOIndex_t inputIdx) const;

    /**
     * @brief Amount of data from a specific input actually used in one computation pass.
     *
     * @param inputIdx Index of the input analysed.
     * @return NbElts_t
     */
    virtual NbElts_t getNbConsumedData(const IOIndex_t inputIdx) const;

    /**
     * @brief Amount of data ready to be used on a specific output.
     *
     * @param outputIdx Index of the output analysed.
     * @return NbElts_t
     */
    virtual NbElts_t getNbProducedData(const IOIndex_t outputIdx) const;

    virtual void updateConsummerProducer();

    virtual void forward();

    virtual void backward();

///////////////////////////////////////////////////////
//        INNER
///////////////////////////////////////////////////////

    std::string type() const {
        return mType;
    }

    inline IOIndex_t nbInputs() const noexcept { return mNbData+mNbAttr; };
    inline IOIndex_t nbData() const noexcept { return mNbData; };
    inline IOIndex_t nbAttr() const noexcept { return mNbAttr; };
    inline IOIndex_t nbOutputs() const noexcept { return mNbOut; };

    static const std::vector<std::string> getInputsName(){
        return {};
    }
    static const std::vector<std::string> getOutputsName(){
        return {};
    }
};
} // namespace Aidge

#endif /* AIDGE_CORE_OPERATOR_OPERATOR_H_ */
